/**
 *  Gaussian regularizer (Chen \& Rosenfeld) 
 *  with a single MU, SIGMA^2 for all features.
 */

public class HomogeneousGaussianRegularizer implements Regularizer
{

public double value(Feature[] features, double[] weights)
{
	double val = 0.0;
	for(int k = weights.length-1; k >= 0; k--)
	{ val += (Math.pow((weights[k]-MU),2)/SIGMA2); }
	return val;
}

public double[] gradient(Feature[] features, double[] weights)
{
	double[] grad = new double[weights.length];
	for(int k = weights.length; k >= 0; k--)
	{ grad[k] = ((2 * (weights[k]-MU))/SIGMA2); }
	return grad;
}

public double sigma2(Feature feature)
{ return SIGMA2; }

}//HomogeneousGaussianRegularizer